-- Addon:   pNameplates
-- Author:  Beermuscles

--[[ ---------------------- ]]--
--[[ Globals            	]]--
--[[ ---------------------- ]]--

default, realm, player, C, addon_loaded = nil, nil, nil, nil, nil
size_update, text_update, font_update   = false, false, false

bkgTex = 'Interface\\Buttons\\WHITE8x8'
barTex = 'Interface\\AddOns\\pNameplates\\Textures\\bar'

ColorString = function( str ) return '|cFF48D1CC'..str..'|r' end
COLOR_LABEL = ColorString( 'p' )..'Nameplates'

--[[ ---------------------- ]]--
--[[ Everything Else    	]]--
--[[ ---------------------- ]]--

local total = 0

local function Copy( t )
	  local u = { }
	  for k, v in pairs( t ) do u[k] = v end
	  return setmetatable( u, getmetatable( t ) )
end

local function GetRaidIconColor( ulx, uly )
	if ( not ulx ) then ulx = -1 end
	if ( not uly ) then uly = -1 end

	if ( uly == 0.00 ) then

		if ( ulx == 0.00 ) then return 1.00, 1.00, 0.00 end -- Star
		if ( ulx == 0.25 ) then return 1.00, 0.50, 0.00 end -- Circle
		if ( ulx == 0.50 ) then return 0.78, 0.30, 1.00 end -- Diamond
		if ( ulx == 0.75 ) then return 0.30, 1.00, 0.30 end -- Triangle

	elseif ( uly == 0.25 ) then

		if ( ulx == 0.00 ) then return 0.60, 0.60, 0.60 end -- Moon
		if ( ulx == 0.25 ) then return 0.20, 0.60, 1.00 end -- Square
		if ( ulx == 0.50 ) then return 1.00, 0.30, 0.30 end -- Cross
		if ( ulx == 0.75 ) then return 1.00, 1.00, 1.00 end -- Skull

	end

	-- default to gray
	return 0.40, 0.40, 0.40
end

local function GetElementLocation( location )
	if ( location == 'top' )     then return 'BOTTOM', 'TOP'   ,  0, 2, 'CENTER' end
	if ( location == 'left' )    then return 'RIGHT' , 'LEFT'  , -2, 0, 'RIGHT'  end
	if ( location == 'right' )   then return 'LEFT'  , 'RIGHT' ,  2, 0, 'LEFT'   end
	if ( location == 'center' )  then return 'CENTER', 'CENTER',  0, 0, 'CENTER' end
	-- if ( location == 'inleft' )  then return 'LEFT'  , 'LEFT'  ,  2, 0, 'THINOUTLINE', 'LEFT'   end
	-- if ( location == 'inright' ) then return 'RIGHT' , 'RIGHT' , -2, 0, 'THINOUTLINE', 'RIGHT'  end

	-- default to top
	return 'BOTTOM', 'TOP', 0, 2, 'CENTER'
end

local function GetApproxHealth( hp )
	if ( not hp ) then return '' end

	local div, tag
	
	if 	   ( hp < 1000 )       then return tostring( hp )
	elseif ( hp < 1000000 )    then div, tag = 1000,       'k'
	elseif ( hp < 1000000000 ) then div, tag = 1000000,    'm'
	else							div, tag = 1000000000, 'b'
	end
	
	return ( tostring( hp / div ) ):match( '%d+\.%d' )..tag
end

local function AbbreviateName( name )
	name_str, len, cur, abbrv = name:GetText(), 0, 0, ''
	
	for token in name_str:gmatch( '[^%s]+' ) do
		cur = token:len() + 1
		len = len + cur
		abbrv = abbrv..token:sub(1, 1)..'. '
	end
	
	len = len - cur
	return ( abbrv:sub( 1, abbrv:len() - 3 )..name_str:sub( len ) )
end

local function GetCheckSum( P )
	local alpha = tostring( P:GetAlpha() )
	alpha = alpha:sub( 1, math.min( alpha:len(), 4 ) )
	local blue = tostring( select( 3, P.old.HB:GetStatusBarColor() ) )
	blue = blue:sub( 1, math.min( blue:len(), 4 ) )
	
	local str = P.old.nam:GetText()..blue..alpha
	local n = str:len()
    local sum = 0
       
    for i = 1, n do sum = 2 * sum + str:byte( i, i ) end
    return sum
end

local function Update_Data( P )
	local HB, CB, old = P.HB, P.CB, P.old
	P.checksum = GetCheckSum( P )
	
	HB.abbv_name = AbbreviateName( old.nam )
	HB.full_name = old.nam:GetText()
	HB.chop_name = old.nam:GetStringWidth() > C.hpw + 10 and C.Abbreviate

	local r, g, b = old.HB:GetStatusBarColor()
	
	HB:SetStatusBarColor( old.HB:GetStatusBarColor() )
	HB:SetMinMaxValues( old.HB:GetMinMaxValues() )

	lvl:SetTextColor( old.lvl:GetTextColor() )
	if ( old.lvl:GetText() ~= nil ) then
		lvl:SetText( old.lvl:GetText()..( old.dgn:IsShown() and '+' or '' ) )
	end
end

local function Update_Size( P, scale )
	if ( not scale ) then scale = 1 end

	P.HB:SetSize( C.hpw * scale, C.hph * scale )
	P.CB:SetSize( C.hpw * scale, C.cbh * scale )
	P.CB.ico:SetSize( C.cbh * scale, C.cbh * scale )

	Update_Data( P )
end

local function Update_Font( P )
	P.HB.hp:SetFont(   C.FontType_Perc,  C.FontSize_Perc,  'THINOUTLINE' )
	P.HB.nam:SetFont(  C.FontType_Name,  C.FontSize_Name,  'THINOUTLINE' )
	P.old.nam:SetFont( C.FontType_Name,  C.FontSize_Name,  'THINOUTLINE' )
	P.HB.lvl:SetFont(  C.FontType_Level, C.FontSize_Level, 'THINOUTLINE' )
	P.CB.spl:SetFont(  C.FontType_Cast,  C.FontSize_Cast,  'THINOUTLINE' )
	P.CB.tim:SetFont(  C.FontType_Cast,  C.FontSize_Cast,  'THINOUTLINE' )
end

local function Update_Text( P )
	local pt, rel, x, y, justify

	pt, rel, x, y, justify = GetElementLocation( C.Location_Perc )
	P.HB.hp:ClearAllPoints()
	P.HB.hp:SetPoint( pt, P.HB, rel, x, y )
	P.HB.hp:SetJustifyH( justify )

	pt, rel, x, y, justify = GetElementLocation( C.Location_Name )
	P.HB.nam:ClearAllPoints()
	P.HB.nam:SetPoint( pt, P.HB, rel, x, y )
	P.HB.nam:SetJustifyH( justify )

	pt, rel, x, y, justify = GetElementLocation( C.Location_Level )
	P.old.bos:ClearAllPoints()
	P.old.bos:SetPoint( pt, P.HB, rel, x, y )
	P.HB.lvl:ClearAllPoints()
	P.HB.lvl:SetPoint( pt, P.HB, rel, x, y )
	P.HB.lvl:SetJustifyH( justify )

	pt, rel, x, y, _ = GetElementLocation( C.Location_Raid )
	P.HB.ico:ClearAllPoints()
	P.HB.ico:SetPoint( pt, P.HB, rel, x, y )
end

local function SetHBData( old, HB )
	HB.nam:SetText( HB.chop_name and HB.abbv_name or HB.full_name )

	local min, max = old.HB:GetMinMaxValues()
	local hp_val = old.HB:GetValue()

	HB:SetMinMaxValues( min, max )
	HB:SetValue( hp_val )

	local hp_string = ''

	if ( C.ShowPercent and ( C.ShowWhenMax or hp_val ~= max ) ) then
		hp_string = 'hp_percent%'
		if ( C.ShowHPValue ) then hp_string = 'hp_value | '..hp_string end

		local value = GetApproxHealth( hp_val )
		local percent = tostring( hp_val / max * 100 ):match( '%d+' )

		hp_string = hp_string:gsub( '(hp_value)',   value )
		hp_string = hp_string:gsub( '(hp_percent)', percent )
	end

	HB.hp:SetText( hp_string )
end

local function SetCBData( CB, value )
	if ( not value ) then value = CB:GetValue() end

	value = tostring( value ):match( '%d+\.%d' )
	CB.tim:SetText( value and value or '0.0' )

	if ( not CB.shl:IsShown() ) then CB:SetStatusBarColor( 1.00, 0.75, 0.00, 1 )
	else                             CB:SetStatusBarColor( 0.75, 0.75, 0.75, 1 ) end
end

local function SetRaidIconStatus( HB )
	HB:SetStatusBarColor( GetRaidIconColor( HB.ico:GetTexCoord() ) )
end

local function GetThreatStatus( old )
	local r, g, b = old.thr:GetVertexColor()
	local dps = C.DPSHealerMode

	if ( old.thr ~= nil and not old.thr:IsShown() ) then
		if ( dps ) then return 0, 1, 0, 1, nil, 'LOW',  1, 2
		else            return 1, 0, 0, 1, 1.2, 'HIGH', 4, 5 end
	end
	if ( g > 0 ) then   return 1, 1, 0, 1, nil, 'MED',  4, 5 end
	if ( r > 0 ) then 
		if ( dps ) then return 1, 0, 0, 1, 1.2, 'HIGH', 4, 5
		else            return 0, 1, 0, 1, nil, 'LOW',  1, 2 end
	end

	-- default to unit's hostility color
	return r, g, b, 1, nil, 'LOW', 1, 2
end

local function Update_Threat( P, HB, old )
	local r, b, g, a, scale, style, x, y = GetThreatStatus( old )

	Update_Size( P, scale )
	HB:SetStatusBarColor( r, b, g, a )

	HB.thr:SetPoint( 'TOPRIGHT',   HB.bdr,  x,  y )
	HB.thr:SetPoint( 'BOTTOMLEFT', HB.bdr, -x, -y )
	HB.thr:SetTexture( 'Interface\\AddOns\\pNameplates\\Textures\\threat_'..style )
	
	if ( not HB.thr:IsVisible() ) then HB.thr:Show() end
end

local function DefaultAppearance( P, HB, old )
	if( HB.thr:IsShown() ) then
		Update_Size( P )
		HB.thr:Hide()
	end

	HB:SetStatusBarColor( old.HB:GetStatusBarColor() )
end

local function ShowOutline( P )
	return C.TargetBorder and UnitExists( 'target' ) and P:GetAlpha() > 0.9
end

local function Update_Main( P )
	-- P:SetAlpha( math.max( 0.75, P:GetAlpha() ) )

	local HB, CB, old = P.HB, P.CB, P.old

	SetHBData( old, HB )
	HB.ico:SetSize( C.RaidIconSize, C.RaidIconSize )

	-- HP Bar Color Filtering via priority
	if     ( select( 3, old.HB:GetStatusBarColor() ) > 0 )            then DefaultAppearance( P, HB, old ) -- tagged or friendly
	elseif ( C.RaidIconColor and HB.ico:IsShown() )                   then SetRaidIconStatus( HB )     -- raid icon
	elseif ( C.ThreatEnabled and UnitAffectingCombat( 'player' ) )    then Update_Threat( P, HB, old ) -- threat style
	else     DefaultAppearance( P, HB, old ) end -- default

	if ( ShowOutline( P ) )  then HB.tgt:Show() else HB.tgt:Hide() end
	if ( old.ovr:IsShown() ) then HB.ovr:Show() else HB.ovr:Hide() end
	if ( old.bos:IsShown() ) then HB.lvl:Hide() else HB.lvl:Show() end

	if ( size_update ) then Update_Size( P ) end -- frame size has changed
	if ( text_update ) then Update_Text( P ) end -- text location has changed
	if ( font_update ) then Update_Font( P ) end -- font size has changed
end

local function SpawnBorder( anchor, edge )
	border = CreateFrame( 'Frame', nil, anchor )
	border:SetPoint( 'TOPLEFT', anchor, 'TOPLEFT', -edge, edge )
	border:SetPoint( 'BOTTOMRIGHT', anchor, 'BOTTOMRIGHT', edge, -edge )
	border:SetBackdrop({ 
		bgFile = nil, 
		edgeFile = bkgTex,
		tile = false, tileSize = 0, edgeSize = 1, 
		insets = { left = edge, right = edge, top = edge, bottom = edge }
	})
	
	return border
end

local function ModPlate( P )
	old_frame, old_name = P:GetChildren()
	
	old = { }
	old.nam = old_name:GetRegions()
	old.HB, old.CB = old_frame:GetChildren()
	old.thr, old.bdr, old.ovr, old.lvl, old.bos, old.ico, old.dgn = old_frame:GetRegions()
	old.CB.bar, old.CB.bdr, old.CB.shl, old.CB.ico, old.CB.spl, old.CB.shw = old.CB:GetRegions()

	old.HB:Hide()
	old.nam:Hide()
	old.bdr:Hide()
	old.lvl:SetWidth( 1 )
	old.ovr:SetTexture( 0, 0, 0, 0 )
	old.thr:SetTexture( 0, 0, 0, 0 )
	old.dgn:SetTexture( 0, 0, 0, 0 )
	old.CB.bar:SetTexture( 0, 0, 0, 0 )
	old.CB.bdr:SetTexture( 0, 0, 0, 0 )
    old.CB.shl:SetTexture( 0, 0, 0, 0 )
	old.CB.shw:SetTexture( 0, 0, 0, 0 )
	
	HB = CreateFrame( 'StatusBar', P:GetName(), P )
	HB:SetPoint( 'BOTTOM', P, 'BOTTOM', 0, 5 )
	HB:SetStatusBarTexture( barTex )
    HB:SetBackdrop({
        bgFile = bkgTex,
        insets = { left = 0, right = 0, top = 0, bottom = 0 }
    })
    HB:SetBackdropColor( 0, 0, 0, 0.6 )
	
	bdr = SpawnBorder( HB, 1 )
	bdr:SetBackdropBorderColor( 0, 0, 0, 0.6 )
	
	tgt = SpawnBorder( bdr, 2 )
	tgt:SetBackdropBorderColor( 1, 1, 1, 1 )

	thr = HB:CreateTexture( nil, 'Texture' )
	thr:SetPoint( 'TOPRIGHT',   bdr,  1,  2 )
    thr:SetPoint( 'BOTTOMLEFT', bdr, -1, -2 )
	
	ovr = HB:CreateTexture( nil, 'Texture' )
	ovr:SetDrawLayer( "OVERLAY", 7 )
    ovr:SetTexture( 1, 1, 1, 0.25 )
	ovr:SetPoint( "TOPLEFT", HB, "TOPLEFT", 0, 0 )
	ovr:SetPoint( "BOTTOMRIGHT", HB, "BOTTOMRIGHT", 0, 0 )
	ovr:Hide()

	nam = HB:CreateFontString( nil )
	nam:SetWordWrap( false )
	nam:SetPoint( 'BOTTOM', HB, 'TOP', 0, 3 ) 
	nam:SetFont( C.FontType, C.FontSize_Name, 'THINOUTLINE' )

	old.nam:SetWordWrap( false )
	old.nam:SetFont( C.FontType, C.FontSize_Name, 'THINOUTLINE' )

	lvl = HB:CreateFontString( nil )
	lvl:SetPoint( 'LEFT', HB, 'RIGHT', 3, 0 ) 
	lvl:SetFont( C.FontType, C.FontSize_Level, 'THINOUTLINE' )

	old.bos:SetPoint( lvl:GetPoint() )

	old.ico:SetPoint( 'RIGHT', HB, 'LEFT', -3, 0 )
	old.ico:SetSize( C.RaidIconSize, C.RaidIconSize )

	hp  = HB:CreateFontString( nil )
	hp:SetPoint( 'CENTER', HB, 'CENTER', 0, 0 )
	hp:SetFont( C.FontType, C.FontSize_Perc, 'THINOUTLINE' )

	HB.bdr = bdr
	HB.tgt = tgt
	HB.thr = thr
	HB.ovr = ovr
	HB.nam = nam
	HB.lvl = lvl
	HB.ico = old.ico
	HB.hp  = hp

	SetHBData( old, HB )

	old.HB:HookScript( 'OnValueChanged',  function() SetHBData( old, HB ) end )
	old.HB:HookScript( 'OnMinMaxChanged', function() SetHBData( old, HB ) end )
	
	CB = old.CB
	CB:ClearAllPoints()
	CB:SetPoint( 'TOPLEFT', HB, 'BOTTOMLEFT', 0, -8 )
	CB:SetStatusBarTexture( barTex )
    CB:SetBackdrop({
        bgFile = bkgTex,
        insets = { left = 0, right = 0, top = 0, bottom = 0 }
    })
    CB:SetBackdropColor( 0, 0, 0, 0.6 )
	
	CB_border = SpawnBorder( CB, 1 )
	CB_border:SetBackdropBorderColor( 0, 0, 0, 0.6 )

	CB.ico:ClearAllPoints()
	CB.ico:SetPoint( 'LEFT', CB, 'RIGHT', 5, 0 )
	CB.ico:SetTexCoord( 0.1, 0.1, 0.1, 0.9, 0.9, 0.1, 0.9, 0.9 )
	CB.ico:SetSize( C.cbh, C.cbh )

	ico_border = SpawnBorder( CB, 1 )
	ico_border:SetBackdropBorderColor( 0, 0, 0, 0.6 )

	CB.spl:ClearAllPoints()
	CB.spl:SetPoint( 'LEFT', CB, 'LEFT', 1, 0 )
	CB.spl:SetFont( C.FontType_Cast, C.FontSize_Cast, 'THINOUTLINE' )

	CB.tim = CB:CreateFontString( nil, 'HIGH' )
	CB.tim:SetPoint( 'RIGHT', CB, 'RIGHT', -1, 0 )
	CB.tim:SetFont( C.FontType_Cast, C.FontSize_Cast, 'THINOUTLINE' )

	CB:Hide()

	CB:HookScript( 'OnValueChanged',   function( self, value )    SetCBData( CB, value ) end )
	CB:HookScript( 'OnMinMaxChanged',  function( self, min, max ) SetCBData( CB, nil )   end )

	P.old = old
	P.HB = HB
	P.CB = CB
	P.mod = true

	Update_Main( P )
	Update_Data( P )
	Update_Size( P )
	Update_Font( P )
	Update_Text( P )
	
	P:HookScript( 'OnShow', Update_Main )
	P:HookScript( 'OnHide', Update_Main )
	P:HookScript( 'OnUpdate', function()
		Update_Main( P )

		if ( not P.checksum or P.checksum ~= GetCheckSum( P ) ) then
			Update_Data( P )
		end
	end)

	P:HookScript( 'OnEvent', function( self, event )
		-- if ( not self:IsVisible() ) then return end

		-- print( event )
		if ( event == 'RAID_TARGET_UPDATE' ) then
			SetRaidIconStatus( self.HB )
		end
	end)
	
	P:RegisterEvent( 'RAID_TARGET_UPDATE' )
end

--[[ ---------------------- ]]--
--[[ Addon Initialization 	]]--
--[[ ---------------------- ]]--

local function Initialize_Defaults()
	default = { }
	
	-- Fonts and Sizing
	default.FontSize_Perc  = 12
	default.FontSize_Name  = 10
	default.FontSize_Cast  = 10
	default.FontSize_Level = 12
	default.FontType_Perc  = 'Fonts\\ARIALN.ttf'
	default.FontType_Name  = 'Fonts\\ARIALN.ttf'
	default.FontType_Cast  = 'Fonts\\ARIALN.ttf'
	default.FontType_Level = 'Fonts\\ARIALN.ttf'
	default.hpw            = 120
	default.hph            = 14
	default.cbw            = 120
	default.cbh            = 10
	default.RaidIconSize   = 24
	default.hp_string      = 'hp_value | hp_percent%'
	
	-- Boolean Variables
	default.Abbreviate     = true
	default.ShowPercent    = true
	default.ShowWhenMax    = true
	default.ShowHPValue    = true
	default.TargetBorder   = true
	default.ThreatEnabled  = true
	default.DPSHealerMode  = true
	default.RaidIconColor  = true
	default.RecolorTagged  = true
	
	-- Text Element Positions
	default.Location_Raid  = 'left'
	default.Location_Perc  = 'center'
	default.Location_Name  = 'top'
	default.Location_Level = 'right'
end

local function pNameplates_Initialize()
	Initialize_Defaults()

	realm = GetRealmName()
 	player = UnitName( 'player' )
 	SetCVar( 'ShowClassColorInNameplate', 1 )

	if ( not config ) 				 then config = { } 				  end
	if ( not config[realm] ) 		 then config[realm] = { } 	 	  end
	if ( not config[realm][player] ) then config[realm][player] = { } end

	C = config[realm][player]
	
	-- check font vars
	if ( not C.FontSize_Perc )  then C.FontSize_Perc  = default.FontSize_Perc  end
	if ( not C.FontSize_Name )  then C.FontSize_Name  = default.FontSize_Name  end
	if ( not C.FontSize_Cast )  then C.FontSize_Cast  = default.FontSize_Cast  end
	if ( not C.FontSize_Level ) then C.FontSize_Level = default.FontSize_Level end
	if ( not C.FontType_Perc )  then C.FontType_Perc  = default.FontType_Perc  end
	if ( not C.FontType_Name )  then C.FontType_Name  = default.FontType_Name  end
	if ( not C.FontType_Cast )  then C.FontType_Cast  = default.FontType_Cast  end
	if ( not C.FontType_Level ) then C.FontType_Level = default.FontType_Level end
	if ( not C.hp_string )      then C.hp_string      = default.hp_string      end

	-- check size vars
	if ( not C.hpw )          then C.hpw 	      = default.hpw 	     end
	if ( not C.hph )          then C.hph 	      = default.hph 	     end
	if ( not C.cbw )          then C.cbw 	      = default.cbw 	     end
	if ( not C.cbh )          then C.cbh          = default.cbh 	     end
	if ( not C.RaidIconSize ) then C.RaidIconSize = default.RaidIconSize end
	
	-- check boolean vars
	if ( C.Abbreviate == nil )    then C.Abbreviate     = default.Abbreviate     end
	if ( C.ShowPercent == nil )   then C.ShowPercent    = default.ShowPercent    end
	if ( C.ShowWhenMax == nil )   then C.ShowWhenMax    = default.ShowWhenMax    end
	if ( C.ShowHPValue == nil )   then C.ShowHPValue    = default.ShowHPValue    end
	if ( C.TargetBorder == nil )  then C.TargetBorder   = default.TargetBorder   end
	if ( C.ThreatEnabled == nil ) then C.ThreatEnabled  = default.ThreatEnabled  end
	if ( C.DPSHealerMode == nil ) then C.DPSHealerMode  = default.DPSHealerMode  end
	if ( C.RaidIconColor == nil ) then C.RaidIconColor  = default.RaidIconColor  end
	if ( C.RecolorTagged == nil ) then C.RecolorTagged  = default.RecolorTagged  end
	
	-- check position vars
	if ( not C.Location_Raid )  then C.Location_Raid  = default.Location_Raid  end
	if ( not C.Location_Perc )  then C.Location_Perc  = default.Location_Perc  end
	if ( not C.Location_Name )  then C.Location_Name  = default.Location_Name  end
	if ( not C.Location_Level ) then C.Location_Level = default.Location_Level end

	addon_loaded = true
	UpdateConfigItems()
end

function pNameplates_OnLoad( self )
	self:RegisterEvent( 'ADDON_LOADED' )
end

function pNameplates_OnEvent( self, event, ... )
	local addon = select( 1, ... )

	if ( addon == 'pNameplates' ) then
		pNameplates_Initialize()
	end
end

function pNameplates_OnUpdate( self, elapsed )
	if ( not addon_loaded ) then return end

	local N = WorldFrame:GetNumChildren()
	-- timer = timer + elapsed
	
	if ( N ~= total or configure ) then
		total = N
		
		for i = 1, N do
			local P = select( i, WorldFrame:GetChildren() )
			if ( P:GetName() and P:GetName():match( 'NamePlate' ) ) then
		  		if ( not P.mod ) then ModPlate( P ) end
		  		if ( configure ) then Update_Main( P ) end
		  	end
		end

		if ( configure ) then
			configure = false
			size_update = false
			font_update = false
			text_update = false
		end 
	end
end